{*************************************************}
{*  Atari Portfolio Serial Port Server           *}
{*                                               *}
{*  (c) 2014 Data M&T Soft.                      *}
{*                                               *}
{*                                               *}
{*************************************************}
{$M 5120,0,0}
{last modified 16.4.2014}
{buglist
}
{todo
-pf check
}
uses Dos;

const

{------------------- 8250 registers --------------------}
RBR= 0;		{ received data - read only -  if DLAB=0 }
THR= 0;		{ transmit data - write only - if DLAB=0 }

DLL= 0;		{ baud generator (LSB) - if DLAB=1 }
DLM= 1;		{ baud generator (MSB) - if DLAB=1 }

IER= 1;		{ interrupt enable - r/w }
IIR= 2;		{ interrupt identification - read only }
LCR= 3;		{ line control register - r/w }
MCR= 4;		{ modem control register - r/w }
LSR= 5;		{ line status - read only }
MSR= 6;		{ modem status register - read only }

{------------------- 8250 register bits ----------------}
IER_RX= 1;	{ intr on received data - priority 2 }
IER_TX= 2;	{ intr on data transmitted - priority 3 }
IER_ERR= 4;	{ intr on receive error (see LSR) - priority 1 }
IER_MS= 8;	{ intr on modem sts change (see MSR) - priority 4 }

IIR_PEND= 1;	{ active intr request }
IIR_MASK= 6;	{ source id mask }
IIR_MS= 0;	{ modem sts change }
IIR_TX= 2;	{ data transmitted }
IIR_RX= 4;	{ data received }
IIR_ERR= 6;	{ receive error }

LCR_5BIT= 0;	{ byte length codes }
LCR_6BIT= 1;
LCR_7BIT= 2;
LCR_8BIT= 3;

LCR_2STOP= 4;	{ two stop bits }
LCR_PTYEN= 8;	{ parity enabled }
LCR_EVPTY= $10;	{ even parity }
LCR_FIXPTY= $20;{ fixed parity }
LCR_INTREN= $40;{ interrupt enabled }
LCR_DLAB= $80;	{ baud gen. access enabled }

MCR_DTR= 1;	{ DTR modem signal }
MCR_RTS= 2;	{ RTS modem signal }
MCR_OUT1= 4;	{ user signal }
MCR_OUT2= 8;	{ user signal - wired to 3 state IRQ driver }
MCR_LOOP= $10;	{ test loop }

LSR_RX= 1;	{ data received  }
LSR_OR= 2;	{ overrun - read once }
LSR_PR= 4;	{ parity error - read once }
LSR_FR= 8;	{ frame error - read once }
LSR_BT= $10;	{ break error - read once }
LSR_BF= $20;	{ tx buffer empty }
LSR_TX= $40;	{ serializer empty }

MSR_CSR_CH= 1;	{ CSR has changed - read once }
MSR_DSR_CH= 2;	{ DSR has changed - read once }
MSR_RI_END= 4;	{ RI ended - read once }
MSR_RLSD_CH= 8;	{ RLSD changed - read once }
MSR_CSR= $10;
MSR_DSR= $20;
MSR_RI= $40;
MSR_RLSD= $80;

DLM_300= 1;	{ 300 Baud }
DLM_HIGH= 0;	{ more than 300 Baud }

DLL_300= $80;	{ 300 Baud }
DLL_600= $c0;	{ 600 Baud }
DLL_1200= $60;	{ 1200 Baud }
DLL_2400= $30;	{ 2400 Baud }
DLL_4800= $18;	{ 4800 Baud }
DLL_9600= $0c;	{ 9600 Baud }
DLL_19200= $06;	{ 19200 Baud }

procedure GotoXY(X,Y:byte);assembler;
  asm
   mov bh,0
   mov dl,X
   mov dh,Y
   mov ah,2
   int $10
  end;

procedure ClrScr;assembler;
 asm
  mov cx,0
  mov dh,7
  mov dl,79
  mov al,0
  mov ah,6
  mov bh,15
  int $10
  mov bh,0
  xor dx,dx
  mov ah,2
  int $10
 end;

 function KeyPressed:boolean;
  var flag:boolean;
  begin
   asm
    mov flag,0
    mov ah,1
    int $16
    jz  @l1
    inc flag
  @l1:
    mov al,flag
    mov @Result,al
   end;
  end;

 function EscPressed:boolean;
  var flag:boolean;
  begin
   asm
    mov flag,0
    mov ah,1
    int $16
    jz  @l1
    mov ah,0
    int $16
    cmp al,27
    jne @l1
    inc flag
  @l1:
    mov al,flag
    mov @Result,al
   end;
  end;

 procedure WriteXY(x,y:byte;s:string);
  begin
   GotoXY(x,y);
   write(s);
  end;

 var GaugeX        : byte;
     GaugeY        : byte;
     GaugeMin      : integer;
     GaugeMax      : integer;
     GaugeScreenPos: byte;

 procedure GaugeUpdate(aProgress:integer);
  var lGSP:integer;
      pStr:string[20];
  begin
   lGSP:=(aProgress*19) div (GaugeMax-GaugeMin);
   if GaugeScreenPos=lGSP then Exit;
   if GaugeScreenPos<lGSP then
    begin
     FillChar(pStr,lGSP-GaugeScreenPos+1,$FE);
     pStr[0]:=chr(lGSP-GaugeScreenPos);
     WriteXY(GaugeX+GaugeScreenPos,GaugeY,pStr);
    end
   else
    begin
     FillChar(pStr,GaugeScreenPos-lGSP+1,32);
     pStr[0]:=chr(GaugeScreenPos-lGSP);
     WriteXY(GaugeX+lGSP,GaugeY,pStr);
    end;
   GaugeScreenPos:=lGSP;
  end;

 procedure GaugeInit(x,y:byte);
  begin
   WriteXY(x,y+1,'0%.....50%.....100%');
   GaugeX:=x;
   GaugeY:=y;
   GaugeMin:=0;
   GaugeMax:=100;
   GaugeScreenPos:=0;
  end;

 var portaddr:word absolute $0000:$0400;

 procedure DisableInterrupts; inline($FA {cli} );
 procedure EnableInterrupts;  inline($FB {sti} );

 function CommOpen:boolean;
  var x: byte;
  begin
   if (port[portaddr+IIR] and $00F8) = 0 then
    begin
     port[portaddr + LCR]:= LCR_DLAB;
     port[portaddr + DLM]:= DLM_HIGH;
     port[portaddr + DLL]:= DLL_9600;
     port[portaddr + LCR]:= LCR_8BIT;
     port[portaddr + IER]:= 0;
     DisableInterrupts;
     x:=port[portaddr + LSR];
     x:=port[portaddr + RBR];
     port[portaddr + MCR]:= 0;
     EnableInterrupts;
     CommOpen:=true;
    end
   else
    CommOpen:=false;
  end;

 procedure CommClose;
  begin
   DisableInterrupts;
   port[portaddr + MCR]:= 0;
   EnableInterrupts;
  end;

 function TransmitByte(x:byte):boolean;
  var Esc: boolean;
  begin
   port[portaddr + MCR]:= MCR_RTS+MCR_DTR;
   Esc:=false;
   while ((port[portaddr + MSR] and MSR_CSR) = 0) and not Esc do
    Esc:=EscPressed;
   while ((port[portaddr + LSR] and LSR_BF) = 0) and not Esc do
    Esc:=EscPressed;
   port[portaddr + MCR]:= 0;
   if not Esc then
    begin
     DisableInterrupts;
     port[portaddr + THR]:= x;
     EnableInterrupts;
    end;
   TransmitByte:=not Esc;
  end;

 function ReceiveByte(var x:byte):boolean;
  var Esc: boolean;
  begin
   Esc:=false;
   while ((port[portaddr + MSR] and MSR_CSR) = 0) and not Esc do
    Esc:=EscPressed;
   port[portaddr + MCR]:= MCR_RTS+MCR_DTR;
   while ((port[portaddr + LSR] and LSR_RX) = 0) and not Esc do
    Esc:=EscPressed;
   port[portaddr + MCR]:= 0;
   if not Esc then
     begin
      DisableInterrupts;
      x:=port[portaddr + RBR];
      EnableInterrupts;
     end;
   ReceiveByte:=not Esc;
  end;

 function Synchronize:boolean;
  const SynStr:string='START';
  var i,x: byte;
  begin
   i:=1;
   repeat
    if not ReceiveByte(x) then
     begin
      Synchronize:=false;
      Exit;
     end;
    if x=Ord(SynStr[i]) then inc(i)
    else i:=1;
   until i>Length(SynStr);
   Synchronize:=true;
  end;

const pfCMD_nocommand    =0;
      pfCMD_setdatetime  =1;
      pfCMD_findfirst    =2;
      pfCMD_findnext     =3;
      pfCMD_openfile     =4;
      pfCMD_copyinternal =5;
      pfCMD_closefile    =6;
      pfCMD_blockread    =7;
      pfCMD_blockwrite   =8;
      pfCMD_setattr      =9;
      pfCMD_delete       =10;
      pfCMD_rmdir        =11;
      pfCMD_rename       =12;
      pfCMD_makedir      =13;
      pfCMD_getdrives    =14;
      pfCMD_getdir       =15;
      pfCMD_setvolumeid  =16;
      pfCMD_getdisksizes =17;
      pfCMD_exec         =18;

      pfMSG_ok           =100;
      pfMSG_error        =101;

type TPFBuffer=array[0..2047] of byte;
     TPFCommand=record
       Command:byte;
       Len:word;
       Buff:TPFBuffer;
      end;

var CMD: TPFCommand;
    Stop: boolean;
    GlobalFile : file;

type TAction=(acListen,acReceive,acTransmit,acProcess);
     TName = array[0..256] of char;

procedure ShowAction(action:TAction);
 begin
  case action of
   acListen   : WriteXY(10,3,'Listening...             ');
   acReceive  : WriteXY(10,3,'Receiving...             ');
   acTransmit : WriteXY(10,3,'Transmiting...           ');
   acProcess  : WriteXY(10,3,'Processing...            ');
  end;
 end;

procedure Init;
 begin
  {pf check}
  ClrScr;
  asm
   mov al,1        {portfolio box}
   mov bh,0
   mov dx,0
   mov cx,$0727
   mov ah,$09
   int $60
  end;
  WriteXY(3,0,'Atari Portfolio Serial Port Server');
  WriteXY(1,1,'Data M&T Soft. (c) 2014 version 1.0');
  WriteXY(1,3,'Action:');
  GaugeInit(10,4);
  WriteXY(5,7,'...to stop server press ESC...');
  GotoXY(0,5);
  Stop:=false;
  {wait for connection}
  if not CommOpen then
   begin
    ClrScr;
    writeln('Cannot initialize serial port.');
    Halt(1);
   end;
 end;

procedure Done;
 begin
  CommClose;
  ClrScr;
 end;

function TransmitMessage:boolean;
 var i:word;
 begin
  with CMD do
   begin
    GaugeMax:=Len+3;
    if not TransmitByte(Command) then
     begin
      TransmitMessage:=false;
      Exit;
     end;
    GaugeUpdate(1);
    if not TransmitByte(Lo(Len)) then
     begin
      TransmitMessage:=false;
      Exit;
     end;
    GaugeUpdate(2);
    if not TransmitByte(Hi(Len)) then
     begin
      TransmitMessage:=false;
      Exit;
     end;
    GaugeUpdate(3);
    for i:=1 to Len do
     if not TransmitByte(Buff[i-1]) then
      begin
       TransmitMessage:=false;
       Exit;
      end
     else
      GaugeUpdate(i+3);
    TransmitMessage:=true;
   end;
 end;

function ReceiveCommand:boolean;
 var i:word;
     LoLength,HiLength:byte;
 begin
  with CMD do
   begin
    if not ReceiveByte(Command) then
     begin
      ReceiveCommand:=false;
      Exit;
     end;
    if not ReceiveByte(LoLength) then
     begin
      ReceiveCommand:=false;
      Exit;
     end;
    if not ReceiveByte(HiLength) then
     begin
      ReceiveCommand:=false;
      Exit;
     end;
    Len:=256*HiLength+LoLength;
    GaugeMax:=Len+3;
    GaugeUpdate(3);
    for i:=1 to Len do
     if not ReceiveByte(Buff[i-1]) then
      begin
       ReceiveCommand:=false;
       Exit;
      end
     else
      GaugeUpdate(i+3);
    ReceiveCommand:=true;
   end;
 end;

{procedures called from 'case' in 'run'}
procedure PFError(errorcode:byte);
 begin
  with CMD do
   begin
    Command:=pfMSG_error;
    Len:=1;
    Buff[0]:=errorcode;
   end;
 end;

procedure PFFindFirst;
var
  SRec: SearchRec;
  Atr : word;
  Path: string;
  i   : byte;
begin
 with CMD do
  begin
   Path:='';
   for i:=1 to Buff[0] do
    Path:=Path+Chr(Buff[i]);
   Atr:=Buff[Len-1]*256+Buff[Len-2];
   FindFirst(Path, Atr, SRec);
  end;
 if DosError=0 then
  with CMD do
   begin
    Command:=pfMSG_ok;
    Len:=SizeOf(SearchRec);
    Move(SRec,Buff,Len);
   end
 else PFError(DosError);
end;

procedure PFFindNext;
var
  SRec: SearchRec;
begin
 with CMD do
  begin
   Move(Buff,SRec,SizeOf(SearchRec));
   FindNext(SRec);
  end;
 if DosError=0 then
  with CMD do
   begin
    Command:=pfMSG_ok;
    Len:=SizeOf(SearchRec);
    Move(SRec,Buff,Len);
   end
 else PFError(DosError);
end;

procedure PFGetDrives;
 var i:byte;
     Drives:string[26];
     flag:boolean;
 begin
  Drives:='';
  for i:=0 to 25 do
   begin
    asm
     mov ah,$0E
     mov dl,i
     int $21
     mov ah,$19
     int $21
     mov dl,0
     mov flag,dl
     cmp al,i
     jne @l1
     inc flag
     @l1:
    end;
     if flag then Drives:=Drives+chr(i+ord('A'));
   end;
  with CMD do
   begin
    Command:=pfMSG_ok;
    Len:=Length(Drives);
    for i:=1 to Len do
     Buff[i-1]:=ord(Drives[i]);
   end
 end;

procedure PFGetDir;
var
  SRec: SearchRec;
  Atr : word;
  Path: string;
  i   : byte;
begin
 with CMD do
  begin
   Command:=pfMSG_ok;
   Path:='';
   for i:=1 to Buff[0] do
    Path:=Path+Chr(Buff[i]);
   Atr:=Buff[Len-1]*256+Buff[Len-2];
   Len:=0;
   FindFirst(Path, Atr, SRec);
   if DosError<>0 then
    begin
     PFError(DosError);
     Exit;
    end;
   while DosError=0 do
    begin
     with SRec do
      begin
       Buff[Len]:=Attr;
       inc(Len);
       Move(Time,Buff[Len],SizeOf(Time));
       inc(Len,SizeOf(Time));
       Move(Size,Buff[Len],SizeOf(Size));
       inc(Len,SizeOf(Size));
       for i:=0 to Length(Name) do
        begin
         Buff[Len]:=ord(Name[i]);
         inc(Len);
        end;
      end;
     FindNext(SRec);
    end;
  end;
end;

type

 TExtendedFCB = Record
    ExtendedFCBflag: Byte;
    Reserved1      : array[1..5] of Byte;
    Attr           : Byte;
    DriveID        : Byte;
    FileName       : array[1..8] of Char;
    FileExt        : array[1..3] of Char;
    CurrentBlockNum: Word;
    RecordSize     : Word;
    FileSize       : LongInt;
    PackedDate     : Word;
    PackedTime     : Word;
    Reserved2      : array[1..8] of Byte;
    CurrentRecNum  : Byte;
    RandomRecNum   : LongInt;
  end;

TVolumeLabel = string[12];

procedure PFSetVolumeId;

var
  OldVolume,NewVolume: TVolumeLabel;
  FCB: TExtendedFCB;
  ErrorFlag: Boolean;
  i:byte;
  SRec: SearchRec;
begin
  for i:=1 to 11 do
   NewVolume[i]:=chr(CMD.Buff[i]);
  NewVolume[0] := Chr(11);
  ErrorFlag := False;
  FindFirst(chr(CMD.Buff[0]+64)+':\????????.???',VolumeID,SRec);
  if DosError=0 then OldVolume:=SRec.Name
                else OldVolume:='';
  if Pos('.', OldVolume) > 0 then Delete(OldVolume, 9, 1);
  if OldVolume<>'' then
   begin
    for i := Length(OldVolume) + 1 to 11 do OldVolume[i] := ' ';
    OldVolume[0] := Chr(11);
    FillChar(FCB, SizeOf(FCB), 0);    { initialize FCB with zeros }
    with FCB do
    begin
      ExtendedFCBflag := $FF;         { always }
      Attr := VolumeID;               { Volume ID attribute }
      DriveID := CMD.Buff[0];         { Drive number }
      Move(OldVolume[1], FileName, 8);  { must enter volume label }
      Move(OldVolume[9], FileExt, 3);
    end;
    asm
      PUSH    DS                      { preserve ds }
      MOV     AX,SS                   { put seg of FCB (ss) in ds }
      MOV     DS,AX
      LEA     DX,FCB                  { put offset of FCB in dx }
      MOV     AX,$1300                { function 13h (delete file) }
      INT     $21                     { invoke int 21h }
      POP     DS                      { restore ds }
      CMP     AL,00h                  { check for success }
      JE      @@End
    @@Error:                          { set flag on error }
      MOV     ErrorFlag,1
    @@End:
    end;
   end;
  if NewVolume[1]<>' ' then
   begin
    FillChar(FCB, SizeOf(FCB), 0);    { initialize FCB with zeros }
    with FCB do
    begin
      ExtendedFCBflag := $FF;         { always }
      Attr := VolumeID;               { Volume ID attribute }
      DriveID := CMD.Buff[0];         { Drive number }
      Move(NewVolume[1], FileName, 8);  { must enter volume label }
      Move(NewVolume[9], FileExt, 3);
    end;
    asm
      PUSH    DS                      { preserve ds }
      MOV     AX,SS                   { put seg of FCB (ss) in ds }
      MOV     DS,AX
      LEA     DX,FCB                  { put offset of FCB in dx }
      MOV     AX,$1600                { function 16h (create file) }
      INT     $21                     { invoke int 21h }
      POP     DS                      { restore ds }
      CMP     AL,00h                  { check for success }
      JE      @@End
    @@Error:                          { set flag on error }
      MOV     ErrorFlag,1
    @@End:
    end;
   end;
  CMD.Command:=pfMSG_ok;
  CMD.Len:=0;
end;

procedure PFStrOp(op: byte);
 var s:string;
 begin
  Move(CMD.Buff[0],s,CMD.Buff[0]+1);
  case op of
   1: MkDir(s);
   2: RmDir(s);
  end;
  if IOResult=0 then
   begin
    CMD.Command:=pfMSG_ok;
    CMD.Len:=0;
   end
  else
   PFError(IOResult);
 end;

procedure PFSetSysDateTime;
 var loy,hiy:byte;
     month,day,hour,min,sec:byte;
 begin
  with CMD do
   begin
    loy  :=Buff[0];
    hiy  :=Buff[1];
    month:=Buff[2];
    day  :=Buff[3];
    hour :=Buff[4];
    min  :=Buff[5];
    sec  :=Buff[6];
   end;
  asm
   mov ah,$2B
   mov ch,hiy
   mov cl,loy
   mov dh,month
   mov dl,day
   int $21
   mov ah,$2D
   mov ch,hour
   mov cl,min
   mov dh,sec
   mov dl,0
   int $21
  end;
  CMD.Command:=pfMSG_ok;
  CMD.Len:=0;
 end;

procedure PFGetDiskSizes;
 var rax,rbx,rcx,rdx:word;
     drive:byte;
 begin
  drive:=CMD.Buff[0];
  asm
   mov ah,$36
   mov dl,drive
   int $21
   mov rax,ax
   mov rbx,bx
   mov rcx,cx
   mov rdx,dx
  end;
  with CMD do
   begin
    Command:=pfMSG_ok;
    Len:=8;
    Move(rax,Buff[0],2);
    Move(rbx,Buff[2],2);
    Move(rcx,Buff[4],2);
    Move(rdx,Buff[6],2);
   end;
 end;

procedure PFExec;
 var CmdLine:string;
     f:text;
 begin
  with CMD do
   begin
    Move(Buff[0],CmdLine,Buff[0]+1);
    Assign(f,'exec.bat');
    Rewrite(f);
    Writeln(f,'@'+CmdLine);
    Writeln(f,'@fs.bat');
    Close(f);
    Stop:=true;
    Command:=pfMSG_ok;
    Len:=0;
   end;
 end;

procedure PFOpenFile;
 var FileName: string;
 begin
  with CMD do
   begin
    Move(Buff[1],FileName,Buff[1]+1);
    Assign(GlobalFile,FileName);
    FileMode:=Buff[0];
    if Buff[0]=0 then Reset(GlobalFile,1)
                 else Rewrite(GlobalFile,1);
    if IOResult=0 then
     begin
      Command:=pfMSG_ok;
      Len:=0;
     end
    else PFError(IOResult);
   end;
 end;

procedure PFBlockRead;
 var NumRead: word;
     Count: word;
 begin
  with CMD do
   begin
    Move(Buff,Count,SizeOf(Count));
    BlockRead(GlobalFile,Buff,Count,NumRead);
    if IOResult=0 then
     begin
      Command:=pfMSG_ok;
      Len:=NumRead;
     end
    else PFError(IOResult);
   end;
 end;

procedure PFBlockWrite;
 var NumWritten: word;
     Count: word;
 begin
  with CMD do
   begin
    Count:=Len;
    BlockWrite(GlobalFile,Buff,Count,NumWritten);
    if IOResult=0 then
     begin
      Command:=pfMSG_ok;
      Len:=2;
      Move(NumWritten,Buff,2);
     end
    else PFError(IOResult);
   end;
 end;

procedure PFCloseFile;
 begin
  Close(GlobalFile);
  with CMD do
   if IOResult=0 then
    begin
     Command:=pfMSG_ok;
     Len:=0;
    end
   else PFError(IOResult);
 end;

procedure PFSetAttr;
 var Atr: word;
     Tm: longint;
 begin
  with CMD do
   begin
    Move(Buff,Atr,SizeOf(Atr));
    SetFAttr(GlobalFile,Atr);
    if Len>2 then
     begin
      Move(Buff[2],Tm,SizeOf(Tm));
      SetFTime(GlobalFile,Tm);
     end;
    if DosError=0 then
     begin
      Command:=pfMSG_ok;
      Len:=0;
     end
    else PFError(DosError);
   end;
 end;


procedure PFRename;
 var OldName,
     NewName: TName;
     ErrorCode: byte;
 begin
  FillChar(OldName,256,0);
  FillChar(NewName,256,0);
  with CMD do
   begin
    Move(Buff[1],OldName,Buff[0]);
    Move(Buff[Buff[0]+2],NewName,Buff[Buff[0]+1]);
    asm
     PUSH    DS
     MOV     AX,SS
     MOV     DS,AX
     LEA     DX,OldName
     MOV     ES,AX
     LEA     DI,NewName
     MOV     AX,$5600
     MOV     ErrorCode,0
     INT     $21
     POP     DS
     JNC     @@END
     MOV     ErrorCode,AL
     @@END:
    end;
    if ErrorCode=0 then
     begin
      Command:=pfMSG_ok;
      Len:=0;
     end
    else PFError(ErrorCode);
   end;
 end;

procedure PFDeleteFile;
 var FileName: TName;
     ErrorCode: byte;
 begin
  FillChar(FileName,256,0);
  with CMD do
   begin
    Move(Buff[1],FileName,Buff[0]);
    asm
     PUSH    DS
     MOV     AX,SS
     MOV     DS,AX
     LEA     DX,FileName
     MOV     AX,$4100
     MOV     ErrorCode,0
     INT     $21
     POP     DS
     JNC     @@END
     MOV     ErrorCode,AL
     @@END:
    end;
    if ErrorCode=0 then
     begin
      Command:=pfMSG_ok;
      Len:=0;
     end
    else PFError(ErrorCode);
   end;
 end;

procedure PFCopyInternal;
 const cCopyBufSize=1024;
 var FileName: string;
     FromF: file;
     NumRead, NumWritten: Word;
     Error: byte;
 begin
  with CMD do
   begin
    Move(Buff[0],FileName,Buff[0]+1);
    Assign(FromF,FileName);
    FileMode:=0;
    Reset(FromF,1);
    if IOResult=0 then
     begin
      Error:=0;
      repeat
       BlockRead(FromF, Buff, cCopyBufSize, NumRead);
       BlockWrite(GlobalFile, Buff, NumRead, NumWritten);
       if (NumRead>0) and (NumRead<>NumWritten) then Error:=255;
      until (NumRead < cCopyBufSize) or (Error>0);
      Close(FromF);
      if Error>0 then
       PFError(Error)
      else
       begin
        Command:=pfMSG_ok;
        Len:=0;
       end;
     end
    else PFError(IOResult);
   end;
 end;
{-------------------------------}

procedure Run;
 begin
  repeat
   repeat
    GaugeUpdate(0);
    ShowAction(acListen);
    if not Synchronize then Exit;
    ShowAction(acReceive);
   until ReceiveCommand;
   ShowAction(acProcess);
   case CMD.Command of
    pfCMD_findfirst    :PFFindFirst;
    pfCMD_findnext     :PFFindNext;
    pfCMD_getdrives    :PFGetDrives;
    pfCMD_getdir       :PFGetDir;
    pfCMD_setvolumeid  :PFSetVolumeId;
    pfCMD_makedir      :PFStrOp(1);
    pfCMD_setdatetime  :PFSetSysDateTime;
    pfCMD_getdisksizes :PFGetDiskSizes;
    pfCMD_exec         :PFExec;
    pfCMD_openfile     :PFOpenFile;
    pfCMD_closefile    :PFCloseFile;
    pfCMD_blockread    :PFBlockRead;
    pfCMD_blockwrite   :PFBlockWrite;
    pfCMD_delete       :PFDeleteFile;
    pfCMD_rmdir        :PFStrOp(2);
    pfCMD_rename       :PFRename;
    pfCMD_setattr      :PFSetAttr;
    pfCMD_copyinternal :PFCopyInternal;
    else
     PFError(0);
   end;
   ShowAction(acTransmit);
   TransmitMessage;
  until EscPressed or Stop;
 end;

 {main}
begin
 {DebugInit;}
 Init;
 Run;
 Done;
 {DebugDone;}
end.
